from __future__ import annotations

from requests import Response
from pydantic import BaseModel

from .api.v1.job import ErrorResponse

def unwrap_or_err(response_model: type[BaseModel], response: Response):
    if not response.ok:
        raise Exception(ErrorResponse.parse_raw(response.text).msg)

    return response_model.parse_raw(response.text)



from __future__ import annotations

from enum import IntEnum
from abc import ABC, abstractmethod
from typing import Callable

class JobState(IntEnum):
    WAITING = 0
    RUNNING = 1
    STOPPED = 2
    FAILED = 3
    DONE = 4
    DOES_NOT_EXIST = 5

class JobStatus:
    status_id: JobState = 0
    status_text: str = ""
    is_done: bool = False
    has_failed: bool = False

class DelegateConfig(ABC):
    def __init__(self, **kwargs):
        """
        Easy way to apply some number of named arguments onto self.
        """

        self.__dict__ = {**self.__dict__, **kwargs}

class Delegate(ABC):
    type: str = ""

    @abstractmethod
    def __init__(self, delegate_config: DelegateConfig):
        """
        Instantiate a new delegate instance.
        """
        pass

    @abstractmethod
    def connect(self) -> bool:
        """
        Connect to the compute instance.

        :returns: True if successful, False if not.
        """
        pass

    @abstractmethod
    def test_connection(self) -> bool:
        """
        Test the delegate's connection to a compute instance.
        """
        pass

    @abstractmethod
    def create_job(self, job_id: str) -> bool:
        """
        Create a new job with the specified ID.

        :returns: True if successful, False if not.
        """
        pass

    @abstractmethod
    def start_job(self, job_id: str, work: Callable, *args, **kwargs) -> bool:
        """
        Start a job with the specified ID.

        :returns: True if successful, False if not.
        """
        pass

    @abstractmethod
    def stop_job(self, job_id: str) -> bool:
        """
        Stop a job with the specified ID.

        :returns: True if successful, False if not.
        """
        pass

    @abstractmethod
    def job_status(self, job_id: str) -> JobStatus:
        """
        Get the status of a job with the specified ID.

        :returns: A JobStatus instance with status details.
        """
        pass

    @abstractmethod
    def job_results(self, job_id: str) -> str:
        """
        The raw results of a job with the specified ID.

        :returns: The results object of the work done.
        """
        pass

    @abstractmethod
    def job_exists(self, job_id: str) -> bool:
        """
        Check if a job with the specified ID exists in the delegate.

        :returns: True if the job exists, False if it does not.
        """
        pass

    @abstractmethod
    def job_complete(self, job_id: str) -> bool:
        """
        Check if a job with the specified ID is complete.

        :returns: True if the job is complete, False if it is not.
        """
        pass




import os
import bz2
import tempfile

import plotly.io as plotlyio

from flask import Blueprint
from flask import make_response

from matplotlib import pyplot

from gillespy2.core import Results

from .apiutils import delegate

from .job import ErrorResponse
from .job import StartJobResponse

v1_memory = Blueprint("V1 Result API Endpoint", __name__, url_prefix="memory/")

@v1_memory.route("/<string:memory_id>/get", methods=["GET"])
def get_memory(memory_id: str):
    if not delegate.job_complete(memory_id):
        return ErrorResponse(msg="A memory with this ID does not yet exist.").json(), 404

    memory_json = delegate.job_results(memory_id)
    compressed_memory = bz2.compress(memory_json.encode())

    response = make_response(compressed_memory)
    response.headers["Content-Encoding"] = "bzip2"

    return response, 200


@v1_memory.route("/<string:memory_id>/exists", methods=["GET"])
def memory_exist(memory_id: str):
    if not delegate.job_complete(memory_id):
        return "False", 404

    return "True", 200

@v1_memory.route("/<string:memory_id>/average_ensemble", methods=["POST"])
def make_average_ensemble(memory_id: str):
    # Make sure we can grab a memory with this ID.
    job_id = f"average_ensemble-{memory_id}"

    if delegate.job_exists(job_id):
        return StartJobResponse(
            job_id=job_id,
            msg="The job has already been started.",
            status=f"/v1/job/{job_id}/status"
        ).json(), 202

    delegate.start_job(job_id, Results.average_ensemble, f"memory://{memory_id}")

    return StartJobResponse(
        job_id=job_id,
        msg="The job has been successfully started.",
        status=f"/v1/job/{job_id}/status"
    ).json(), 202

@v1_memory.route("/<string:memory_id>/plot", methods=["GET"])
def make_plot(memory_id: str):
    # Make sure we can grab a memory with this ID.
    if not delegate.job_complete(memory_id):
        status = delegate.job_status(memory_id)
        return status.status_text, 404

    # Grab the memory.
    memory: Results = Results.from_json(delegate.job_results(memory_id))

    # Swap the pyplot backend so the Results#plot call wont try to write to a GUI.
    pyplot.switch_backend("template")

    # Write the plot as a .png to the tempfile.
    _, temp = tempfile.mkstemp(suffix=".png")
    memory.plot(save_png=temp)

    # Read the contents of the temp file and cleanup.
    with open(temp, "rb") as infile:
        plot_bytes = infile.read()

    os.remove(temp)

    compressed_plot = bz2.compress(plot_bytes)

    response = make_response(compressed_plot)
    response.headers["Content-Encoding"] = "bzip2"

    return response, 200

@v1_memory.route("/<string:memory_id>/plotplotly", methods=["GET"])
def make_plotplotly(memory_id: str):
    # Ensure that a memory with this ID exists.
    if not delegate.job_complete(memory_id):
        return "Something broke", 404

    # Grab the memorys.
    memory: Results = delegate.job_results(memory_id)

    # Plot the figure without rendering, returning the backing datastructure.
    memorys_plot = memory.plotplotly(return_plotly_figure=True)

    # Generate plot JSON, encode, and compress.
    plot_json = plotlyio.to_json(memorys_plot)
    compressed_plot = bz2.compress(plot_json.encode())

    # Create and return the HTTP response.
    response = make_response(compressed_plot)
    response.headers["Content-Encoding"] = "bzip2"

    return response, 200



from __future__ import annotations

import os

from stochss_compute.api.cache import CacheProvider, CacheProviderConfig

class SimpleDiskCacheConfig(CacheProviderConfig):
    root_dir = "sd-cache/"

class SimpleDiskCache(CacheProvider):
    def __init__(self, config: SimpleDiskCacheConfig):
        self.config = config

    def put(self, id: str, value: str):
        # Write value to a file with the name `id`. If a file already exists, replace it.
        with open(os.path.join(self.config.root_dir, id), "w+") as outfile:
            outfile.write(value)

    def get(self, id: str):
        # Read the file in as a string, return the contents.
        with open(os.path.join(self.config.root_dir, id), "r") as infile:
            return infile.read()

    def exists(self, id: str) -> bool:
        # Check to see if the file exists on disk.
        return os.path.isfile(os.path.join(self.config.root_dir, id))



from __future__ import annotations

from abc import ABC
from abc import abstractmethod

class CacheProviderConfig(ABC):
    def __init__(self, **kwargs):
        """
        Easy way to apply some number of named arguments onto self.
        """

        self.__dict__ = {**self.__dict__, **kwargs}

class CacheProvider(ABC):
    def __init__(self, config: CacheProviderConfig):
        pass

    @abstractmethod
    def put(id: str, value):
        pass

    @abstractmethod
    def get(id: str):
        pass

    @abstractmethod
    def exists(id: str) -> bool:
        pass



import os

from typing import List
from typing import Dict

from ..apiutils import delegate
from ..job import JobStatusResponse
from ..dataclass import ErrorResponse

import gillespy2.core

from flask import request
from flask import Blueprint

from pydantic import BaseModel
from pydantic import ValidationError

model_endpoint = Blueprint("V1 GillesPy2 Model API Endpoint", __name__, url_prefix="/model")

class Model(gillespy2.core.Model):
    @classmethod
    def __serialize__(cls, value: "Model") -> str:
        return value.to_json()

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value) -> "Model":
        if isinstance(value, str):
            return cls(Model.from_json(value))

        return value

class ModelRunRequest(BaseModel):
    model: Model
    args: List = []
    kwargs: Dict = {}

    class Config:
        json_encoders = {gillespy2.core.Model: lambda model: Model.__serialize__(model)}

@model_endpoint.route("/run", methods=["POST"])
def run():
    # Attempt to parse request data.
    try:
        run_request = ModelRunRequest.parse_raw(request.json)

    except ValidationError as e:
        return ErrorResponse(msg=f"Invalid request data: '{e}'").json(), 400

    model = run_request.model.name
    model_id = f"{model.get_json_hash()}"
    number_trajectories = int(run_request.kwargs.pop("number_of_trajectories", 1))

    print(run_request.kwargs)
    print(run_request.args)

    import hashlib
    model_id = hashlib.md5(f"{model.get_json_hash()}_{run_request.kwargs}_{run_request.args}".encode()).hexdigest()

    if "solver" in run_request.kwargs:
        from pydoc import locate
        run_request.kwargs["solver"] = locate(run_request.kwargs["solver"])

    # Build a list of job key values which will need to be run.
    keys = [f"{model_id}-trajectory_{i}" for i in range(number_trajectories)]

    # Run each trajectory and save in a dataset.
    # Warning: Here be possible threading issues. Needs investigation.
    dependencies = delegate.client.map(gillespy2.core.Model.run, [model] * number_trajectories, **run_request.kwargs, key=keys)

    cache_dir = delegate.cache_provider.config.root_dir
    os.makedirs(cache_dir, exist_ok=True)

    def join_results(results):
        # Save the dependent trajectories in memory.
        from distributed import get_client
        client = get_client()
        client.publish_dataset(dependencies, name=f"{model_id}-trajectories", override=True)

        data = []

        for i, result in enumerate(results):
            with open(os.path.join(cache_dir, keys[i]), "w+") as outfile:
                outfile.write(result.to_json())

            data = data + result.data

        from gillespy2.core import Results
        result_json = Results(data).to_json()

        with open(os.path.join(cache_dir, job_id), "w+") as outfile:
            outfile.write(result_json)

        return result_json

    job_id = f"{model_id}-run_{number_trajectories}"
    delegate.start_job(job_id, join_results, dependencies)

    # Return the status of the currently running job.
    job_status = delegate.job_status(job_id)
    return JobStatusResponse(
        job_id=job_id,
        status_id=job_status.status_id,
        status_msg=job_status.status_text,
        is_complete=job_status.is_done,
        has_failed=job_status.has_failed
    ).json(), 201




from flask import Blueprint

from .model import model_endpoint
from .results import results_endpoint

v1_gillespy2 = Blueprint("V1 GillesPy2 API Endpoints", __name__, url_prefix="/gillespy2")
v1_gillespy2.register_blueprint(model_endpoint)
v1_gillespy2.register_blueprint(results_endpoint)



import bz2
import sys
import json

import plotly.io as plotlyio

from gillespy2.core import Model
from gillespy2.core import Results

from .remote_utils import unwrap_or_err

from .compute_server import Endpoint
from .compute_server import ComputeServer

from .remote_results import RemoteResults

from .api.v1.gillespy2.model import ModelRunRequest

from .api.v1.job import (
    StartJobRequest,
    StartJobResponse,
    JobStatusResponse,
    JobStopResponse,
    ErrorResponse
)

class RemoteSimulation():
    @classmethod
    def on(cls, server: ComputeServer):
        """
        Specify a stochss-compute server that subsequent simulations will be run on.

        :param server: A ComputeServer instance which contains connection details
            of the stochss-compute instance to run simulations on.
        :type server: ComputeServer
        """

        return RemoteSimulation(server)

    def __init__(self, server: ComputeServer):
        self.server = server

    def with_model(self, model: Model):
        """
        Specify the GillesPy2 Model that will be run in subquent simulations.

        :param model: The GillesPy2 Model to simulate.
        :type model: Model
        """

        self.model = model

        return self

    def run(self, **params) -> RemoteResults:
        """
        Simulate the Model on the target ComputeServer, returning the results once complete.

        :param **params: Arguments to pass directly to the Model#run call on the server.
        
        :returns: Results
        """
    
        if "solver" in params:
            params["solver"] = f"{params['solver'].__module__}.{params['solver'].__qualname__}"

        start_request = ModelRunRequest(model=self.model, kwargs=params)
        start_response = unwrap_or_err(JobStatusResponse, self.server.post(Endpoint.GILLESPY2_MODEL, sub="/run", request=start_request))

        remote_results = RemoteResults(result_id=start_response.job_id, server=self.server)
        return remote_results

    def test_plot(self):
        plot_response = self.server.get(Endpoint.RESULT, f"/{self.model.get_json_hash()}/plot")
        # plot_request = requests.get(f"{self.server.address}/api/v1/result/{self.model.get_json_hash()}/plot")

        print(f"Plot size: {sys.getsizeof(plot_response.content)}")
        plot_json = bz2.decompress(plot_response.content).decode()

        with open("test", "w") as outfile:
            outfile.write(plot_json)

        plot = plotlyio.from_json(plot_json)
        plot.show()

    def __get_job_status(self, status_url: str) -> JobStatusResponse:
        status_response  = self.server.get(Endpoint.JOB, status_url)
        # status_raw = requests.get(status_url)

        if status_response.status_code != 200:
            error = ErrorResponse.parse_raw(status_response.text)
            raise Exception(error)

        return JobStatusResponse.parse_raw(status_response.text)




from gillespy2.core import (
    Event,
    Model,
    Species,
    Reaction,
    RateRule,
    Parameter,
    EventTrigger,
    AssignmentRule,
    EventAssignment,
	FunctionDefinition
)

import numpy as np


class VilarOscillator(Model):
    def __init__(self, parameter_values=None):
        Model.__init__(self, name="VilarOscillator")
        self.volume = 1

        # Parameters
        alphaA = Parameter(name="alphaA", expression=50)
        alphaA_prime = Parameter(name="alphaA_prime", expression=500)
        alphaR = Parameter(name="alphaR", expression=0.01)
        alphaR_prime = Parameter(name="alphaR_prime", expression=50)
        betaA = Parameter(name="betaA", expression=50)
        betaR = Parameter(name="betaR", expression=5)
        deltaMA = Parameter(name="deltaMA", expression=10)
        deltaMR = Parameter(name="deltaMR", expression=0.5)
        deltaA = Parameter(name="deltaA", expression=1)
        deltaR = Parameter(name="deltaR", expression=0.2)
        gammaA = Parameter(name="gammaA", expression=1)
        gammaR = Parameter(name="gammaR", expression=1)
        gammaC = Parameter(name="gammaC", expression=2)
        thetaA = Parameter(name="thetaA", expression=50)
        thetaR = Parameter(name="thetaR", expression=100)
        
        self.add_parameter([alphaA, alphaA_prime, alphaR, alphaR_prime, betaA, betaR,
                            deltaMA, deltaMR, deltaA, deltaR, gammaA, gammaR, gammaC,
                            thetaA, thetaR])

        # Species
        Da = Species(name="Da", initial_value=1, mode="discrete")
        Da_prime = Species(name="Da_prime", initial_value=0, mode="discrete")
        Ma = Species(name="Ma", initial_value=0, mode="discrete")
        Dr = Species(name="Dr", initial_value=1, mode="discrete")
        Dr_prime = Species(name="Dr_prime", initial_value=0, mode="discrete")
        Mr = Species(name="Mr", initial_value=0, mode="discrete")
        C = Species(name="C", initial_value=0, mode="discrete")
        A = Species(name="A", initial_value=0, mode="discrete")
        R = Species(name="R", initial_value=0, mode="discrete")
        
        self.add_species([Da, Da_prime, Ma, Dr, Dr_prime, Mr, C, A, R])

        # Reactions
        r1 = Reaction(name="r1", reactants={'A': 1, 'R': 1}, products={'C': 1}, rate="gammaC")
        r2 = Reaction(name="r2", reactants={'A': 1}, products={}, rate="deltaA")
        r3 = Reaction(name="r3", reactants={'C': 1}, products={'R': 1}, rate="deltaA")
        r4 = Reaction(name="r4", reactants={'R': 1}, products={}, rate="deltaR")
        r5 = Reaction(name="r5", reactants={'A': 1, 'Da': 1}, products={'Da_prime': 1}, rate="gammaA")
        r6 = Reaction(name="r6", reactants={'Da_prime': 1}, products={'A': 1, 'Da': 1}, rate="thetaA")
        r7 = Reaction(name="r7", reactants={'Da': 1}, products={'Da': 1, 'Ma': 1}, rate="alphaA")
        r8 = Reaction(name="r8", reactants={'Da_prime': 1}, products={'Da_prime': 1, 'Ma': 1}, rate="alphaA_prime")
        r9 = Reaction(name="r9", reactants={'Ma': 1}, products={}, rate="deltaMA")
        r10 = Reaction(name="r10", reactants={'Ma': 1}, products={'A': 1, 'Ma': 1}, rate="betaA")
        r11 = Reaction(name="r11", reactants={'A': 1, 'Dr': 1}, products={'Dr_prime': 1}, rate="gammaR")
        r12 = Reaction(name="r12", reactants={'Dr_prime': 1}, products={'A': 1, 'Dr': 1}, rate="thetaR")
        r13 = Reaction(name="r13", reactants={'Dr': 1}, products={'Dr': 1, 'Mr': 1}, rate="alphaR")
        r14 = Reaction(name="r14", reactants={'Dr_prime': 1}, products={'Dr_prime': 1, 'Mr': 1}, rate="alphaR_prime")
        r15 = Reaction(name="r15", reactants={'Mr': 1}, products={}, rate="deltaMR")
        r16 = Reaction(name="r16", reactants={'Mr': 1}, products={'Mr': 1, 'R': 1}, rate="betaR")
        
        self.add_reaction([r1, r2, r3, r4, r5, r6, r7, r8, r9,
                           r10, r11, r12, r13, r14, r15, r16])

        # Timespan
        self.timespan(np.linspace(0,200,201))


class Dimerization(Model):
    def __init__(self, parameter_values=None):
        # First call the gillespy2.Model initializer.
        Model.__init__(self, name="Dimerization")

        # Define parameters for the rates of creation and dissociation.
        k_c = Parameter(name='k_c', expression=0.005)
        k_d = Parameter(name='k_d', expression=0.08)
        self.add_parameter([k_c, k_d])

        # Define variables for the molecular species representing M and D.
        m = Species(name='monomer', initial_value=30)
        d = Species(name='dimer',   initial_value=0)
        self.add_species([m, d])

        # Define the reactions representing the process.  In GillesPy2,
        # the list of reactants and products for a Reaction object are each a
        # Python dictionary in which the dictionary keys are Species objects
        # and the values are stoichiometries of the species in the reaction.
        r_creation = Reaction(name="r_creation", rate=k_c,
                                            reactants={m:2}, products={d:1})
        r_dissociation = Reaction(name="r_dissociation", rate=k_d,
                                            reactants={d:1}, products={m:2})
        self.add_reaction([r_creation, r_dissociation])

        # Set the timespan for the simulation.
        self.timespan(np.linspace(0, 100, 101))


class Trichloroethylene(Model):
    """
    UNCA iGEM 2017 Metabolic Channel.
    """

    def __init__(self, parameter_values=None):
        # initialize Model
        Model.__init__(self, name="Trichloroethylene")

        # Species
        A = Species(name='TCE', initial_value=300)
        B = Species(name='Epoxide', initial_value=120)
        C = Species(name='Dichloracatate', initial_value=0)
        D = Species(name='LossOfOneCL', initial_value=0)
        E = Species(name='Glyoxylate', initial_value=0)
        F = Species(name='Output', initial_value=0)
        self.add_species([A, B, C, D, E, F])

        # Parameters
        K1 = Parameter(name='K1', expression=0.00045 * 0.000025)
        K2 = Parameter(name='K2', expression=14)
        K3 = Parameter(name='K3', expression=0.033)
        K4 = Parameter(name='K4', expression=500 * 0.0001)
        K5 = Parameter(name='K5', expression=500 * 0.0001)
        self.add_parameter([K1, K2, K3, K4, K5])

        # Reactions
        J1 = Reaction(name="J1", reactants={A: 1}, products={B: 1}, rate=K2)

        J2 = Reaction(name="J2", reactants={B: 1}, products={C: 1}, rate=K3)

        J3 = Reaction(name="J3", reactants={C: 1}, products={D: 1}, rate=K4)

        J4 = Reaction(name="J4", reactants={D: 1}, products={E: 1}, rate=K4)

        J5 = Reaction(name="J5", reactants={E: 1}, products={F: 1}, rate=K5)
        self.add_reaction([J1, J2, J3, J4, J5])
        self.timespan(np.linspace(0, 10000, 100))


class LacOperon(Model):
    """
    Heath LS, Cao Y. Problem Solving Handbook in Computational Biology and Bioinformatics. Springer; 2014.
    """

    def __init__(self, parameter_values=None):
        # initialize Model
        Model.__init__(self, name="LacOperon")
        
        # Species
        s1 = Species(name='MR', initial_value=0)
        s2 = Species(name='R', initial_value=0)
        s3 = Species(name='R2', initial_value=0)
        s4 = Species(name='O', initial_value=1)
        s5 = Species(name='R2O', initial_value=0)
        s6 = Species(name='I', initial_value=0)
        s7 = Species(name='Iex', initial_value=0)
        s8 = Species(name='I2R2', initial_value=0)
        s9 = Species(name='MY', initial_value=0)
        s10 = Species(name='Y', initial_value=0)
        s11 = Species(name='YIex', initial_value=0)
        s12 = Species(name='Ytot', initial_value=0)
        
        self.add_species(
            [s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12])

        # Parameters
        k1 = Parameter(name='k1', expression=0.111)
        k2 = Parameter(name='k2', expression=15.0)
        k3 = Parameter(name='k3', expression=103.8)
        k4 = Parameter(name='k4', expression=0.001)
        k5 = Parameter(name='k5', expression=1992.7)
        k6 = Parameter(name='k6', expression=2.40)
        k7 = Parameter(name='k7', expression=1.293e-6)
        k8 = Parameter(name='k8', expression=12.0)
        k9 = Parameter(name='k9', expression=1.293e-6)
        k10 = Parameter(name='k10', expression=9963.2)
        k11 = Parameter(name='k11', expression=0.50)
        k12 = Parameter(name='k12', expression=0.010)
        k13 = Parameter(name='k13', expression=30)
        k14 = Parameter(name='k14', expression=0.249)
        k15 = Parameter(name='k15', expression=0.10)
        k16 = Parameter(name='k16', expression=60000)
        k17 = Parameter(name='k17', expression=0.920)
        k18 = Parameter(name='k18', expression=0.920)
        k19 = Parameter(name='k19', expression=0.462)
        k20 = Parameter(name='k20', expression=0.462)
        k21 = Parameter(name='k21', expression=0.20)
        k22 = Parameter(name='k22', expression=0.20)
        k23 = Parameter(name='k23', expression=0.20)
        k24 = Parameter(name='k24', expression=0.20)
        k25 = Parameter(name='k25', expression=0.20)
        self.add_parameter(
            [k1, k2, k3, k4, k5, k6, k7, k8, k9, k10, k11, k12, k13, k14, k15, k16, k17, k18, k19, k20, k21, k22, k23, k24, k25])

        # Reactions
        j1 = Reaction(name="j1", reactants={s12: 1}, products={s1: 1}, rate=k1)
        j2 = Reaction(name="j2", reactants={s1: 1}, products={s1: 1, s2: 1}, rate=k2)
        j3 = Reaction(name="j3", reactants={s2: 2}, products={s3: 1}, rate=k3)
        j4 = Reaction(name="j4", reactants={s3: 1}, products={s2: 2}, rate=k4)
        j5 = Reaction(name="j5", reactants={s3: 1, s4: 1}, products={s5: 1}, rate=k5)
        j6 = Reaction(name="j6", reactants={s5: 1}, products={s3: 1, s4: 1}, rate=k6)
        j7 = Reaction(name="j7", reactants={s6: 2, s3: 1}, products={s8: 1}, propensity_function='((3e-7)/((8e-16)*(6.0221367e14))**(2))*(R2)*(I)*(I-1)')
        j8 = Reaction(name="j8", reactants={s8: 1}, products={s6: 2, s3: 1}, rate=k8)
        j9 = Reaction(name="j9", reactants={s6: 2, s5: 1}, products={s8: 1, s4: 1}, propensity_function='((3e-7)/((8e-16)*(6.0221367e14))**(2))*(R2O)*(I)*(I-1)')
        j10 = Reaction(name="j10", reactants={s8: 1, s4: 1}, products={s6: 2, s5: 1}, rate=k10)
        j11 = Reaction(name="j11", reactants={s4: 1}, products={s4: 1, s9: 1}, rate=k11)
        j12 = Reaction(name="j12", reactants={s5: 1}, products={s5: 1, s9: 1}, rate=k12)
        j13 = Reaction(name="j13", reactants={s9: 1}, products={s9: 1, s10: 1}, rate=k13)
        j14 = Reaction(name="j14", reactants={s10: 1, s7: 1}, products={s11: 1}, rate=k14)
        j15 = Reaction(name="j15", reactants={s11: 1}, products={s10: 1, s7: 1}, rate=k15)
        j16 = Reaction(name="j16", reactants={s11: 1}, products={s10: 1, s6: 1}, rate=k16)
        j17 = Reaction(name="j17", reactants={s7: 1}, products={s6: 1}, rate=k17)
        j18 = Reaction(name="j18", reactants={s6: 1}, products={s7: 1}, rate=k18)
        j19 = Reaction(name="j19", reactants={s1: 1}, products={s12: 1}, rate=k19)
        j20 = Reaction(name="j20", reactants={s9: 1}, products={s12: 1}, rate=k20)
        j21 = Reaction(name="j21", reactants={s2: 1}, products={s12: 1}, rate=k21)
        j22 = Reaction(name="j22", reactants={s3: 1}, products={s12: 1}, rate=k22)
        j23 = Reaction(name="j23", reactants={s10: 1}, products={s12: 1}, rate=k23)
        j24 = Reaction(name="j24", reactants={s11: 1}, products={s6: 1}, rate=k24)
        j25 = Reaction(name="j25", reactants={s8: 1}, products={s6: 2}, rate=k25)
        self.add_reaction(
            [j1, j2, j3, j4, j5, j6, j7, j8, j9, j10, j11, j12, j13, j14, j15, j16, j17, j18, j19, j20, j21, j22, j23, j24, j25])
        self.timespan(np.linspace(0, 100, 10))


class Schlogl(Model):
    """
    Schlogl F. Chemical reaction models for non-equilibrium phase transitions. Zeitschrift for Physik.
    1972;253: 147–161. doi:10.1007/bf01379769
    """

    def __init__(self, parameter_values=None):
        # initialize Model
        Model.__init__(self, name="Schlogl")

        # Species
        s1 = Species(name='A', initial_value=300)
        s2 = Species(name='B', initial_value=300)
        s3 = Species(name='C', initial_value=300)
        s4 = Species(name='X', initial_value=300)

        self.add_species([s1, s2, s3, s4])

        k1 = Parameter(name='k1', expression=1)
        k2 = Parameter(name='k2', expression=1)

        self.add_parameter([k1, k2])

        j1 = Reaction(name="j1", reactants={s1: 1, s4: 1}, products={s4: 2.0}, rate=k1)
        j2 = Reaction(name="j2", reactants={s2: 1, s4: 1}, products={s3: 1}, rate=k2)

        self.add_reaction([j1, j2])
        self.timespan(np.linspace(0, 100000, 100))


class MichaelisMenten(Model):
    def __init__(self, parameter_values=None):
        # initialize Model
        Model.__init__(self, name="Michaelis_Menten")

        # parameters
        rate1 = Parameter(name='rate1', expression=0.0017)
        rate2 = Parameter(name='rate2', expression=0.5)
        rate3 = Parameter(name='rate3', expression=0.1)
        self.add_parameter([rate1, rate2, rate3])

        # Species
        A = Species(name='A', initial_value=301)
        B = Species(name='B', initial_value=120)
        C = Species(name='C', initial_value=0)
        D = Species(name='D', initial_value=0)
        self.add_species([A, B, C, D])

        # reactions
        r1 = Reaction(name="r1", reactants={A: 1, B: 1}, products={C: 1}, rate=rate1)

        r2 = Reaction(name="r2", reactants={C: 1}, products={A: 1, B: 1}, rate=rate2)

        r3 = Reaction(name="r3", reactants={C: 1}, products={B: 1, D: 1}, rate=rate3)
        self.add_reaction([r1, r2, r3])
        self.timespan(np.linspace(0, 100, 101))


class ToggleSwitch(Model):
    """
    Gardner et al. Nature (1999)Construction of a genetic toggle switch in Escherichia coli
    (Transcription from 
    """

    def __init__(self, parameter_values=None):
        # Initialize the model.
        Model.__init__(self, name="Toggle_Switch")
        # Species
        A = Species(name='A', initial_value=10)
        B = Species(name='B', initial_value=10)
        self.add_species([A, B])
        # Parameters
        alpha1 = Parameter(name='alpha1', expression=10)
        alpha2 = Parameter(name='alpha2', expression=10)
        beta = Parameter(name='beta', expression=2)
        gamma = Parameter(name='gamma', expression=2)
        mu = Parameter(name='mu', expression=1)
        self.add_parameter([alpha1, alpha2, beta, gamma, mu])
        # Reactions
        self.add_reaction(Reaction(name="cu", reactants={}, products={'A': 1}, propensity_function="alpha1/(1+pow(B, beta))"))
        self.add_reaction(Reaction(name="cv", reactants={}, products={'B': 1}, propensity_function="alpha2/(1+pow(A, gamma))"))
        self.add_reaction(Reaction(name="du", reactants={'A': 1}, products={}, rate=self.listOfParameters["mu"]))
        self.add_reaction(Reaction(name="dv", reactants={'B': 1}, products={}, rate=self.listOfParameters["mu"]))
        self.timespan(np.linspace(0, 250, 251))


class Example(Model):
    """
    This is a simple example for mass-action degradation of species S.
    """

    def __init__(self, parameter_values=None):
        # Initialize the model.
        Model.__init__(self, name="Example")
        # Species
        S = Species(name='Sp', initial_value=100)
        self.add_species([S])
        # Parameters
        k1 = Parameter(name='k1', expression=3.0)
        self.add_parameter([k1])
        # Reactions
        rxn1 = Reaction(name='S degradation', reactants={S: 1}, products={}, rate=k1)
        self.add_reaction([rxn1])
        self.timespan(np.linspace(0, 20, 101))


class ExampleNoTspan(Model):
    """
    This is a simple example for mass-action degradation of species S.
    """

    def __init__(self, parameter_values=None):
        # Initialize the model.
        Model.__init__(self, name="Example")
        # Species
        S = Species(name='Sp', initial_value=100)
        self.add_species([S])
        # Parameters
        k1 = Parameter(name='k1', expression=3.0)
        self.add_parameter([k1])
        # Reactions
        rxn1 = Reaction(name='S degradation', reactants={S: 1}, products={}, rate=k1)
        self.add_reaction([rxn1])


class Tyson2StateOscillator(Model):
    """
    Here, as a test case, we run a simple two-state oscillator (Novak & Tyson
    2008) as an example of a stochastic reaction system.
    """

    def __init__(self, parameter_values=None):
        Model.__init__(self, name="tyson-2-state", volume=300.0)

        # Species
        X = Species(name='X', initial_value=int(0.65609071 * 300.0))
        Y = Species(name='Y', initial_value=int(0.85088331 * 300.0))
        self.add_species([X, Y])

        P = Parameter(name='p', expression=2.0)
        kt = Parameter(name='kt', expression=20.0)
        kd = Parameter(name='kd', expression=1.0)
        a0 = Parameter(name='a0', expression=0.005)
        a1 = Parameter(name='a1', expression=0.05)
        a2 = Parameter(name='a2', expression=0.1)
        kdx = Parameter(name='kdx', expression=1.0)
        self.add_parameter([P, kt, kd, a0, a1, a2, kdx])

        # creation of X:
        rxn1 = Reaction(name='X production', reactants={}, products={X: 1},
                        propensity_function='300 * 1.0 / (1.0 + (Y * Y / (300 * 300)))')

        # degradadation of X:
        rxn2 = Reaction(name='X degradation', reactants={X: 1}, products={}, rate=kdx)

        # creation of Y:
        rxn3 = Reaction(name='Y production', reactants={X: 1}, products={X: 1, Y: 1}, rate=kt)

        # degradation of Y:
        rxn4 = Reaction(name='Y degradation', reactants={Y: 1}, products={}, rate=kd)

        # nonlinear Y term:
        rxn5 = Reaction(name='Y nonlin', reactants={Y: 1}, products={}, 
                propensity_function='Y / a0 + a1 * (Y / 300) + a2 * Y * Y / (300 * 300)')

        self.add_reaction([rxn1, rxn2, rxn3, rxn4, rxn5])
        self.timespan(np.linspace(0, 100, 101))

# Oregonator system
# http://www.scholarpedia.org/article/Oregonator
class Oregonator(Model):

     def __init__(self, parameter_values = None):

          # Superclass initialization
          Model.__init__(self, name = "Oregonator")
          
          # Species
          F = Species(name = "F", initial_value = 2)
          A = Species(name = "A", initial_value = 250)
          B = Species(name = "B", initial_value = 500)
          C = Species(name = "C", initial_value = 1000)
          P = Species(name = "P", initial_value = 0)
          self.add_species([F, A, B, C, P])

          # Parameters (rates)
          k1 = Parameter(name = "k1", expression = 2.0)
          k2 = Parameter(name = "k2", expression = 0.1)
          k3 = Parameter(name = "k3", expression = 104)
          k4 = Parameter(name = "k4", expression = 4e-7)
          k5 = Parameter(name = "k5", expression = 26.0)
          self.add_parameter([k1, k2, k3, k4, k5])
          
          # Reactions
          reaction1 = Reaction(name = "reaction1", 
                                         reactants = {B: 1, F: 1}, 
                                         products = {A: 1, F: 1}, 
                                         rate = k1)
          reaction2 = Reaction(name = "reaction2", 
                                         reactants = {A: 1, B: 1}, 
                                         products = {P: 1}, 
                                         rate = k2)
          reaction3 = Reaction(name = "reaction3", 
                                         reactants = {A: 1, F: 1}, 
                                         products = {A: 2, C: 1, F: 1}, 
                                         rate = k3)
          reaction4 = Reaction(name = "reaction4", 
                                         reactants = {A: 2}, 
                                         products = {P: 1}, 
                                         rate = k4)
          reaction5 = Reaction(name = "reaction5", 
                                         reactants = {C: 1, F: 1}, 
                                         products = {B: 1, F: 1}, 
                                         rate = k5)
          self.add_reaction([reaction1, reaction2, reaction3, reaction4, reaction5])
          
          # Set timespan of model
          self.timespan(np.linspace(0, 5, 501))


class Degradation(Model):
    def __init__(self, model_name="Degradation"):
        super().__init__(name=model_name)
        
        A = Species(name="A", initial_value=2000)
        self.add_species(A)
        
        k1 = Parameter(name="k1", expression="0.1")
        self.add_parameter(k1)
        
        r1 = Reaction(name="r1", rate="k1", reactants={"A":1})
        self.add_reaction(r1)
        self.timespan(np.arange(0, 151))


class RobustModel(Model):
    def __init__(self, parameter_values=None):
        Model.__init__(self, name="test1")
        self.volume = 1

        # Parameters
        self.add_parameter(Parameter(name="k1", expression="0.5"))
        self.add_parameter(Parameter(name="k2", expression="3.2e-15"))

        # Variables
        self.add_species(Species(name="s1", initial_value=8000, mode="continuous"))
        self.add_species(Species(name="s2", initial_value=0, mode="continuous"))

        # Reactions
        self.add_reaction(Reaction(name="r1", reactants={'s1': 1}, products={'s2': 1}, rate=self.listOfParameters["k2"]))
        self.add_reaction(Reaction(name="r2", reactants={'s1': 1}, products={}, rate=self.listOfParameters["k1"]))
        self.add_reaction(Reaction(name="r3", reactants={'s1': 2}, products={'s1': 3}, propensity_function="sin(1)"))

        # Event Triggers
        e1_trig = EventTrigger(expression="t > 50", initial_value=True, persistent=True)

        # Event Assignments
        e1_assign_1 = EventAssignment(variable="s1", expression="2000")

        # Events
        self.add_event(Event(name="e1", trigger=e1_trig, assignments=[e1_assign_1], delay="t - 40", priority="0", use_values_from_trigger_time=True))

        # Rate Rules
        self.add_rate_rule(RateRule(name="rr1", formula="sin(0.5)", variable="s1"))

        # Assignment Rules
        self.add_assignment_rule(AssignmentRule(name="rr2", formula="8000/(s1+1)", variable="s2"))

        # Function Definitions
        self.add_function_definition(FunctionDefinition(name="multiply", function="x * y", args=['x', 'y']))

        # Timespan
        self.timespan(np.arange(0, 20, 0.05))



import time
import unittest
import tempfile

from threading import Thread

from tests.gillespy2_models import LacOperon
from tests.gillespy2_models import MichaelisMenten

from stochss_compute import api

from stochss_compute import ComputeServer
from stochss_compute import RemoteSimulation
from stochss_compute.api.delegate import JobState
from stochss_compute.api.delegate import DaskDelegateConfig

from distributed import Client
from distributed import Future
from distributed import LocalCluster

class EnsembleTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.cache_dir = tempfile.TemporaryDirectory(suffix="stochss-compute_")
        cls.worker_dir = tempfile.TemporaryDirectory(suffix="stochss-compute_")

        # Set up the Dask client and stochss-compute instance.
        cls.cluster = LocalCluster("127.0.0.1:9797", local_directory=cls.worker_dir.name)
        cls.client = Client(address=cls.cluster.scheduler_address, set_as_default=False)

        host, port = cls.client.scheduler.address.replace("tcp://", "").split(":")

        delegate_config = DaskDelegateConfig(dask_cluster_address=host, dask_cluster_port=port)

        # If set, change cache root_dir to minimize conflicts.
        if hasattr(delegate_config.cache_provider.config, "root_dir"):
            delegate_config.cache_provider.config.root_dir = cls.cache_dir.name

        cls.api_process = Thread(daemon=False, target=api.start_api, kwargs=dict(host="127.0.0.1", port=29681, delegate_config=delegate_config))
        cls.api_process.start()

        cls.compute_server = ComputeServer(host="127.0.0.1", port=29681)

        time.sleep(3)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cluster.close()
        cls.client.close()

        cls.cache_dir.cleanup()
        cls.worker_dir.cleanup()

    def test_run_model_consistency(self):
        """ Test to ensure the API and Cluster return consistent job statuses. """
        
        model = MichaelisMenten()
        model.name = "test_run_model_consistency_MichaelisMenten"

        job = RemoteSimulation.on(self.compute_server).with_model(model).run()

        # Assert that the job hasn't failed yet, and that it does exist on the scheduler.
        assert(job.status().status_id != JobState.FAILED)

        # Wait for the job to complete.
        job.wait()

        # Assert that the API reports status_id as JobState.DONE.
        assert(job.status().status_id == JobState.DONE)

        # Assert that results can be retrieved from the API. This will raise an exception if the request fails.
        Future(job.result_id, client=self.client).result(timeout=5)

    @unittest.skip("Currently broken due to invalid job stop behavior. See issue #61 for more details.")
    def test_job_stop(self):
        """ Test to ensure that running jobs can be stopped. """

        job = RemoteSimulation.on(self.compute_server).with_model(LacOperon()).run()
        time.sleep(5)

        # Stop the job and assert that the API and Cluster report it as removed.
        job.cancel()

        # Assert that the stopped job can no longer be resolved.
        self.assertRaises(Exception, job.resolve())

    @unittest.skip("Broken due to invalid internal result resolution behavior. See issue #62 for more details.")
    def test_job_cache(self):
        """ Test the cache behavior for completed jobs. """

        model = MichaelisMenten()
        model.namee = "test_job_cache_MichaelisMenten"

        job = RemoteSimulation.on(self.compute_server).with_model(model).run()

        # Wait for the job to complete and get the results.
        results = job.resolve().to_json()

        # Assert that results can still be resolved even if the job was cancelled.
        Future(job.result_id, client=self.client).cancel()
        assert(results == job.resolve().to_json())

        # Assert that job results can be resolved from the cache.
        self.client.unpublish_dataset(job.result_id)
        assert(results == job.resolve().to_json())



# =============================================================================
# @file    __version__.py
# @brief   stochss-compute version info
# @license Please see the file named LICENSE.md in parent directory
# @website https://github.com/StochSS/stochss-compute
# =============================================================================


__version__      = "0.9.5"

__title__        = "stochss-compute"
__description__  = "A compute delegation package for the StochSS family of stochastic simulators"
__url__          = "https://github.com/StochSS/stochss-compute"
__download_url__ = "https://pypi.org/project/stochss-compute/#files"
__author__       = "See CONTRIBUTORS.md"
__email__        = "briandrawert@gmail.com"
__license__      = "GPLv3"
__copyright__    = "Copyright (C) 2017-2022"



from flask import g
from flask import current_app

from werkzeug.local import LocalProxy

from ..delegate import Delegate

def get_delegate():
    if "delegate" in g:
        return g.delegate

    delegate_config = current_app.config["DELEGATE_CONFIG"]
    delegate_type = current_app.config["DELEGATE_TYPE"]

    # # May be None
    # kube_cluster = current_app.config["KUBE_CLUSTER"]
    
    delegate = delegate_type(delegate_config)

    if False in (delegate.connect(), delegate.test_connection()):
        raise Exception("Delegate connection failed.")

    return delegate

delegate: Delegate = LocalProxy(get_delegate)



# Note: how to do a PyPI release
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Run the following commands:
#
#   python3 setup.py sdist bdist_wheel
#   twine upload dist/*
#
# =============================================================================

from os import path
from setuptools import setup, find_packages


# The following reads the variables without doing an "import handprint",
# because the latter will cause the python execution environment to fail if
# any dependencies are not already installed -- negating most of the reason
# we're using setup() in the first place.  This code avoids eval, for security.

SETUP_DIR = path.dirname(path.abspath(__file__))

with open(path.join(SETUP_DIR, "README.md"), "r", errors="ignore") as f:
    readme = f.read()

version = {}
with open(path.join(SETUP_DIR, "stochss_compute/__version__.py")) as f:
    text = f.read().rstrip().splitlines()
    vars = [line for line in text if line.startswith("__") and "=" in line]
    for v in vars:
        setting = v.split("=")
        version[setting[0].strip()] = setting[1].strip().replace("\"", "")

setup(name=             version["__title__"],
      version=          version["__version__"],
      description=      version["__description__"],
      author=           version["__author__"],
      author_email=     version["__email__"],
      url=              version["__url__"],
      licence=          version["__license__"],
      packages=         find_packages("."),
      long_description= readme,
      long_description_content_type= "text/markdown",
      classifiers=      [
          "Development Status :: 5 - Production/Stable",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Chemistry",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Medical Science Apps.",
          "Intended Audience :: Science/Research"
      ],

)



import requests

from enum import Enum

from requests.exceptions import ConnectionError
from time import sleep

from pydantic import BaseModel

class Endpoint(Enum):
    JOB = 1
    RESULT = 2
    GILLESPY2_MODEL = 3
    GILLESPY2_RESULTS = 4

class ComputeServer():
    def __init__(self, host, port: int = 80):
        self.host = host
        self.port = port

        self.address = f"http://{host}:{port}/api/v1"

        self.job_api = f"{self.address}/job"
        self.memory_api = f"{self.address}/memory"
        self.gillespy2_model_api = f"{self.address}/gillespy2/model"
        self.gillespy2_results_api = f"{self.address}/gillespy2/results"

        self.endpoints = {
            Endpoint.JOB: self.job_api,
            Endpoint.RESULT: self.memory_api,
            Endpoint.GILLESPY2_MODEL: self.gillespy2_model_api,
            Endpoint.GILLESPY2_RESULTS: self.gillespy2_results_api
        }

    def get(self, endpoint: Endpoint, sub: str) -> requests.Response:
        url = f"{self.endpoints[endpoint]}{sub}"
        print(f"[GET] {url}")
        return requests.get(f"{self.endpoints[endpoint]}{sub}")

    def post(self, endpoint: Endpoint, sub: str, request: BaseModel = None) -> requests.Response:
        url = f"{self.endpoints[endpoint]}{sub}"
        retry = 1
        sec = 3
        while retry <= 3:
            try:
                if request is None:
                    print(f"[POST] {url}")
                    return requests.post(url)
                print(f"[{type(request).__name__}] {url}")
                return requests.post(url, json=request.json())

            except ConnectionError as ce:
                print(f"Connection refused by server. Retrying in {sec} seconds....")
                sleep(sec)
                retry += 1
                sec *= retry
            
            except Exception as e:
                print(f"Unknown error: {e}. Retrying in {sec} seconds....")
                sleep(sec)
                retry += 1
                sec *= retry




import sys
import bz2
import time
import inspect

from functools import partial

import plotly.io as plotlyio

from IPython.display import Image

from gillespy2.core import Model
from gillespy2.core import Results

from .compute_server import Endpoint
from .compute_server import ComputeServer

from .remote_utils import unwrap_or_err

from .api.v1.job import ErrorResponse
from .api.v1.job import StartJobResponse
from .api.v1.job import JobStatusResponse
from .api.v1.gillespy2.results import PlotPlotlyRequest
from .api.v1.gillespy2.results import AverageEnsembleRequest

class RemoteResults(Results):
    def __init__(self, result_id: str, server: ComputeServer):
        Results.__new__(Results)

        self.result_id = result_id
        self.server = server
        self.local_results = None

    def __getattribute__(self, name):
        if hasattr(RemoteResults, f"hook_{name}"):
            name = f"hook_{name}"

        return object.__getattribute__(self, name)

    def __poll_job_status(self) -> bool:
        # Request the status of a running job.
        status_response = self.server.get(Endpoint.JOB, f"/{self.result_id}/status")

        if not status_response.ok:
            raise Exception(ErrorResponse.parse_raw(status_response.text).msg)

        # Parse the body of the response into a JobStatusResponse object.
        status = JobStatusResponse.parse_raw(status_response.text)
        print(status.status_msg)
        print(status.status_id)

        return status.is_complete

    def __local_hook(self, target: str = "", *args, **kwargs):
        print(f"Calling proxy for: '{target}'...")
        if self.local_results is None:
            self.local_results = self.resolve()

        return getattr(self.local_results, target)(*args, **kwargs)

    def hook_average_ensemble(self, *args, **kwargs):
        status_response = self.server.get(Endpoint.JOB, f"/{self.result_id}/status")
        status: JobStatusResponse = unwrap_or_err(JobStatusResponse, status_response)

        if status.has_failed:
            raise Exception("Something broke, not sure. Oh well.")

        start_request = AverageEnsembleRequest(result_id=self.result_id)
        start_response: JobStatusResponse = unwrap_or_err(JobStatusResponse, self.server.post(Endpoint.GILLESPY2_RESULTS, "/average_ensemble", request=start_request))

        if start_response.has_failed:
            raise Exception("Something broke.")

        import copy
        test = copy.deepcopy(self)
        test.result_id = start_response.job_id
        test.data = None
        print(test)

        return test

    def hook_plot(self, *args, **kwargs):
        plot_response = self.server.get(Endpoint.RESULT, f"/{self.result_id}/plot")

        if plot_response.status_code == 404:
            raise Exception(plot_response.text)

        print(f"Plot size: {sys.getsizeof(plot_response.content)}")
        plot_image = bz2.decompress(plot_response.content)
        print(f"Expanded to: {sys.getsizeof(plot_image)}")

        return Image(data=plot_image, format="png")

    def hook_plotplotly(self, *args, **kwargs):
        plot_request = PlotPlotlyRequest(result_id=self.result_id)
        plot_response: JobStatusResponse = unwrap_or_err(JobStatusResponse, self.server.post(Endpoint.GILLESPY2_RESULTS, "/plotplotly", request=plot_request))
        
        if plot_response.has_failed:
            raise Exception("Failed to create plotly figure.")

        print(f"Plot size: {sys.getsizeof(plot_response.content)}")
        plot_json = bz2.decompress(plot_response.content).decode()
        print(f"Expanded to: {sys.getsizeof(plot_json)}")

        with open("test", "w") as outfile:
            outfile.write(plot_json)

        plot = plotlyio.from_json(plot_json)
        plot.show()

    def status(self) -> JobStatusResponse:
        """
        Get the status of the remote job.

        :returns: JobStatusResponse
        """

        status_response = self.server.get(Endpoint.JOB, f"/{self.result_id}/status")
        status: JobStatusResponse = unwrap_or_err(JobStatusResponse, status_response)

        return status

    def wait(self):
        """
        Wait for the job to finish.
        """

        while not self.__poll_job_status():
            time.sleep(5)

    def resolve(self) -> Results:
        """
        Finish the remote job and return the results.
        This function will block until the remote job is complete.

        :returns: Results
        """

        # Poll the job status until it finishes.
        while not self.__poll_job_status():
            time.sleep(5)

        # Request the results of the finished job.
        results_response = self.server.get(Endpoint.RESULT, f"/{self.result_id}/get")

        if not results_response.ok:
            raise Exception(ErrorResponse.parse_raw(results_response.text).msg)

        print(f"Results size: {sys.getsizeof(results_response.content)}")
        results_json = bz2.decompress(results_response.content).decode()
        print(f"Expanded to: {sys.getsizeof(results_json)}")

        # Parse and return the response body into a valid Results object.
        results = Results.from_json(results_json)

        return results

    def cancel(self):
        """
        Cancels the remote job.
        """
        stop_response = self.server.post(Endpoint.JOB, f"/{self.result_id}/stop")




from flask import Blueprint

from .job import v1_job
from .memory import v1_memory
from .gillespy2 import v1_gillespy2

v1_api = Blueprint("stochss-compute API V1", __name__, url_prefix="/api/v1")

v1_api.register_blueprint(v1_job)
v1_api.register_blueprint(v1_memory)
v1_api.register_blueprint(v1_gillespy2)

# print("V1 API has been initialized.")



from __future__ import annotations

import os

from typing import Callable

from distributed import Future, LocalCluster
from distributed import Client
from distributed import get_client
from distributed.scheduler import TaskState


from .delegate import Delegate
from .delegate import JobState
from .delegate import JobStatus
from .delegate import DelegateConfig

from stochss_compute.api.cache import CacheProvider
from stochss_compute.api.cache import SimpleDiskCache
from stochss_compute.api.cache import SimpleDiskCacheConfig

class DaskDelegateConfig(DelegateConfig):

    dask_cluster_port = 8786
    dask_cluster_address = "localhost"

    dask_kwargs = None

    cache_provider: type[CacheProvider] = SimpleDiskCache(SimpleDiskCacheConfig())


class DaskDelegate(Delegate):
    type: str = "dask"

    def __init__(self, delegate_config: DaskDelegateConfig):
        super()

        self.delegate_config = delegate_config
        self.cache_provider = self.delegate_config.cache_provider

        # Attempt to load the global Dask client.
        try:
            self.client = get_client()

        except ValueError as _:

            if self.delegate_config.dask_kwargs is not None:
                dask_cluster = LocalCluster(**self.delegate_config.dask_kwargs)
                self.client = Client(dask_cluster)
            else:
                self.client = Client(f"{self.delegate_config.dask_cluster_address}:{self.delegate_config.dask_cluster_port}")
            
        print(f"Connected to Dask Scheduler:\n{self.client}")

        # Setup functions to be run on the schedule.
        def __scheduler_job_exists(dask_scheduler, job_id: str) -> bool:
            return job_id in dask_scheduler.tasks

        def __scheduler_job_state(dask_scheduler, job_id: str) -> TaskState:
            return dask_scheduler.tasks[job_id].state

        self.scheduler_job_exists = __scheduler_job_exists
        self.scheduler_job_state = __scheduler_job_state

    def __job_state(self, job_id: str) -> TaskState:
        return self.client.run_on_scheduler(self.scheduler_job_state, job_id=job_id)

    def connect(self) -> bool:
        # No need to connect.
        return True

    def test_connection(self) -> bool:
        # Shim this out until I figure out a good way to test a Dask and Redis connection.
        return True

    def create_job(self, job_id: str) -> bool:
        # No concept of creating a job.
        return True

    def start_job(self, job_id: str, work: Callable, *args, **kwargs) -> bool:
        if self.job_exists(job_id) or self.job_complete(job_id):
            return False

        # Parse and replace instances of the internal `result://` proxy protocol.
        # In short, this allows for callees to reference an in-progress or remote job without needing direct access.
        function_args = [(self.client.get_dataset(arg.replace("result://", "")) if isinstance(arg, str) and arg.startswith("result://") else arg) for arg in args]

        # Create a job to run the desired function.
        job_future: Future = self.client.submit(work, *function_args, **kwargs, key=job_id, pure=False)

        # Start additional cache job which depends on the results of the previous.
        cache_future: Future = self.client.submit(self.cache_provider.put, *[job_id, job_future], pure=False)

        # Publish the job as a dataset to maintain state across requests.
        self.client.publish_dataset(job_future, name=job_id, override=True)
        self.client.publish_dataset(cache_future, override=True)

        return True

    def stop_job(self, job_id: str) -> bool:
        if not self.job_exists(job_id):
            return False

        try:
            # Iterate through the dependencies of this job.
            dependencies = self.client.run_on_scheduler(lambda dask_scheduler: [(state.key) for state in dask_scheduler.tasks[id].dependencies])

            # Filter out any weak depenencies. Strong dependencies are suffixed with "/" and the name of the job.
            dependencies = [(dependency) for dependency in dependencies if dependency.replace(id, "").startswith("/")]

            futures = [(Future(key)) for key in dependencies]
            futures.append(Future(job_id))
        except KeyError:
            # do nothing if no dependencies
            pass

        self.client.cancel(Future(job_id))
        self.client.unpublish_dataset(job_id)

        # Hacky fix -- Simulation processes continue executing EVEN IF the parent task is killed.
        def hacky():
            os.system("pkill -f 'Simulation.out'")

        self.client.run(hacky, nanny=True)

        return True

    def job_status(self, job_id: str) -> JobStatus:
        # If the job is complete (results exist as a dataset or in the vault).
        if self.job_complete(job_id):
            status = JobStatus()
            status.status_id = JobState.DONE
            status.status_text = "The job is complete."
            status.has_failed = False
            status.is_done = True

            return status

        # If the job doesn't exist.
        if not self.job_exists(job_id):
            status = JobStatus()
            status.status_id = JobState.DOES_NOT_EXIST
            status.status_text = f"A job with job_id: '{job_id}' does not exist."
            status.has_failed = True
            status.is_done = False

            return status

        status_mapping = {
            "released": (JobState.STOPPED, "The job is known but not actively computing or in memory."),
            "waiting": (JobState.WAITING, "The job is ready to be computed but is waiting for dependencies to arrive in worker memory."),
            "no-worker": (JobState.WAITING, "The job is ready to be computed but is waiting for a worker to become available."),
            "processing": (JobState.RUNNING, "The job is running."),
            "memory": (JobState.DONE, "The job is done and is being held in memory."),
            "erred": (JobState.FAILED, "The job has failed."),
            "done": (JobState.DONE, "The job is done and has been cached / stored on disk.")
        }

        # Grab the task state from the scheduler.
        future_status = self.__job_state(job_id)

        status = JobStatus()
        status.status_id = status_mapping[future_status][0]
        status.status_text = status_mapping[future_status][1]

        status.is_done = status.status_id is JobState.DONE
        status.has_failed = status.status_id is JobState.FAILED

        return status

    def job_results(self, job_id: str):
        # The results of this job may exist on the client dataset.
        if job_id in self.client.datasets:
            print("[DEBUG] Getting results from dataset.")
            return self.client.get_dataset(name=job_id).result()

        # If the results are not in the cache, raise an exception.
        if not self.cache_provider.exists(job_id):
            raise Exception(f"Result with ID '{job_id}' does not exist in the cache.")

        return self.cache_provider.get(job_id)

    def job_complete(self, job_id: str) -> bool:
        # Finished job results must exist within the cache for it to be considered 'done'.
        return self.cache_provider.exists(job_id)

    def job_exists(self, job_id: str) -> bool:
        # Check if the job exists in the scheduler.
        return self.client.run_on_scheduler(self.scheduler_job_exists, job_id=job_id)

    def get_remote_dependency(self, dependency_id: str):
        # Check to see if the job exists as a dataset.
        dependency = self.client.get_dataset(name=dependency_id)

        if dependency is not None:
            return dependency

        raise Exception("Something broke, dependency does not exist within distributed memory.")



import json

from gillespy2.core import Model
from pydantic import BaseModel

from flask import request
from flask import Blueprint

from .apiutils import delegate

class StartJobRequest(BaseModel):
    job_id: str
    model: str
    args: str
    kwargs: str

class StartJobResponse(BaseModel):
    job_id: str
    msg: str
    status: str

class JobStatusResponse(BaseModel):
    job_id: str
    status_id: int
    status_msg: str
    is_complete: bool
    has_failed: bool

class JobStopResponse(BaseModel):
    job_id: str
    msg: str
    success: bool

class ErrorResponse(BaseModel):
    msg: str

v1_job = Blueprint("V1 Job API Endpoint", __name__, url_prefix="/job")

@v1_job.route("/start", methods=["POST"])
def start_job():
    request_obj = StartJobRequest.parse_raw(request.json)

    if delegate.job_complete(request_obj.job_id):
        return ErrorResponse(
            msg=f"The job with id '{request_obj.job_id} has already completed."
        ).json(), 400

    if delegate.job_exists(request_obj.job_id):
        return StartJobResponse(
            job_id=request_obj.job_id,
            msg="The job has already been started.",
            status=f"/v1/job/{request_obj.job_id}/status"
        ).json(), 202

    model = Model.from_json(request_obj.model)
    kwargs = json.loads(request_obj.kwargs)

    from gillespy2.solvers import SSACSolver
    kwargs["solver"] = SSACSolver

    if not "number_of_trajectories" in kwargs:
        delegate.start_job(request_obj.job_id, Model.run, model, **kwargs)

    else:
        trajectories = kwargs["number_of_trajectories"]
        del kwargs["number_of_trajectories"]

        # Hacky, but it works for now.
        keys = [f"{request_obj.job_id}/trajectory_{i}" for i in range(0, trajectories)]
        delegate.client.scatter([model, kwargs])

        dependencies = delegate.client.map(Model.run, [model] * trajectories, **kwargs, key=keys)

        def test_job(*args, **kwargs):
            data = []

            for result in args:
                data = data + result.data

            from gillespy2.core import Results
            return Results(data)

        delegate.start_job(request_obj.job_id, test_job, *dependencies)

    return StartJobResponse(
        job_id=request_obj.job_id,
        msg="The job has been successfully started.",
        status=f"/v1/job/{request_obj.job_id}/status"
    ).json(), 202

@v1_job.route("/<string:job_id>/status")
def job_status(job_id: str):
    # Check if the job has finished.
    if not delegate.job_exists(job_id) and not delegate.job_complete(job_id):
        return ErrorResponse(
            msg=f"A job with id '{job_id}' does not exist."
        ).json(), 404

    job_status = delegate.job_status(job_id)

    return JobStatusResponse(
        job_id=job_id,
        status_id=job_status.status_id,
        status_msg=job_status.status_text,
        is_complete=job_status.is_done,
        has_failed=job_status.has_failed
    ).json(), 200

@v1_job.route("/<string:job_id>/results")
def job_results(job_id: str):
    if not delegate.job_complete(job_id):
        return ErrorResponse(
            msg=f"The job with id {job_id} is not yet complete."
        ).json(), 400

    job_results = delegate.job_results(job_id)
    return job_results.to_json(), 200

@v1_job.route("/<string:job_id>/stop", methods=["POST"])
def job_stop(job_id: str):
    if not delegate.job_exists(job_id):
        return ErrorResponse(
            msg=f"A job with id {job_id} does not exist."
        ).json(), 404

    if not delegate.stop_job(job_id):
        return JobStopResponse(
            job_id=job_id,
            msg=f"Failed to stop job with id '{job_id}'.",
            success=False
        ).json(), 500

    return JobStopResponse(
        job_id=job_id,
        msg=f"Job with id '{job_id}' has been stopped.",
        success=True
    ).json(), 200




from __future__ import annotations

from flask import Flask

from .v1 import v1_api
from .delegate import Delegate
from .delegate import DelegateConfig

from .delegate.dask_delegate import DaskDelegate
from .delegate.dask_delegate import DaskDelegateConfig

from .cache import CacheProvider
from .cache import CacheProviderConfig
from .cache import SimpleDiskCache
from .cache import SimpleDiskCacheConfig

from dask_kubernetes import KubeCluster
import os

def start_api(
        host: str, 
        port: int, 
        debug: bool = False, 
        delegate_type: type[Delegate] = DaskDelegate,
        delegate_config: DelegateConfig = DaskDelegateConfig(),
        # kube: bool = False,
        **kwargs):
    
    """
    Start the REST API with the following arguments.

    :param host: The address to listen on. This may change depending on the use case.
        For example, this may need to be set to '0.0.0.0' if run in a container.
    :type host: str

    :param port: The port to listen on.
    :type port: int

    :param delegate_type: The type of delegate to be used to run simulation jobs. Defaults
        to DaskDelegate.
    :type delegate_type: type[Delegate]

    :param delegate_config: Delegate-specific configuration options. These will be
        passed directly to the Delegate instance. If None then the default config
        will be used.
    :type delegate_config: DelegateConfig

    :param kwargs: Additional arguments to be passed to Flask#run().
    :type kwargs: dict
    """

    # Instantiate the Flask instance on the API and register any required blueprints
    flask = Flask("stochss-compute REST API")
    flask.config.update(
        JSONIFY_PRETTYPRINT_REGULAR=True,
        DELEGATE_TYPE=delegate_type,
        DELEGATE_CONFIG=delegate_config,
    )

    # # Will return None if not being run using the kubernetes manifest where this environment variable is initialized
    # kube_dask_worker_spec = os.environ.get("WORKER_SPEC_PATH")

    # if kube_dask_worker_spec is not None:
    #     kube_cluster = KubeCluster(pod_template=kube_dask_worker_spec, n_workers=1)
    #     if debug:
    #         print("Dask KubeCluster created:")
    #         print(kube_cluster)
    #     flask.config.update(
    #         KUBE_CLUSTER=kube_cluster
    #     )

    flask.register_blueprint(v1_api)

    # use_reloader=False (prevents two KubeClusters from being created)
    # Start the REST API.
    flask.run(host=host, port=port, debug=debug, use_reloader=False, **kwargs)



from typing import Any
from typing import List
from typing import Dict

from ..apiutils import delegate
from ..job import JobStatusResponse
from ..dataclass import ErrorResponse

from gillespy2.core import Results

from flask import request
from flask import Blueprint

from pydantic import BaseModel
from pydantic import ValidationError

results_endpoint = Blueprint("V1 GillesPy2 Results API Endpoint", __name__, url_prefix="/results")

class PlotPlotlyRequest(BaseModel):
    result_id: str
    args: List[str] = []
    kwargs: Dict[str, Any] = {}

class AverageEnsembleRequest(BaseModel):
    result_id: str

@results_endpoint.route("/plotplotly", methods=["POST"])
def plotplotly():
    # Attempt to parse incoming request data.
    try:
        plot_request = PlotPlotlyRequest.parse_raw(request.json)

    except ValidationError as e:
        return ErrorResponse(msg=f"Invalid request data: '{e}'").json(), 400

    job_id = f"{plot_request.result_id}-plotplotly"

    # Ensure that the parent job (result_id) exists or has saved results.
    if not delegate.job_exists(plot_request.result_id) and not delegate.job_complete(plot_request.result_id):
        return ErrorResponse(msg=f"No parent job with ID '{plot_request.result_id}' exists.").json(), 400

    # If a job or results object does not exist for this ID, start it.
    if not delegate.job_exists(job_id) and not delegate.job_complete(job_id):
        # Hacky fix to ensure that the output plot is of the correct format.
        def make_plotplotly(results_json: str, *args, **kwargs) -> str:
            import plotly.io as plotlyio

            results = Results.from_json(results_json)
            plot = Results.plotplotly(results, return_plotly_figure=True, **kwargs)
            plot_json = plotlyio.to_json(plot)

            return plot_json
                
        delegate.start_job(job_id, make_plotplotly, f"result://{plot_request.result_id}", *plot_request.args, **plot_request.kwargs)

    # Return the status of the currently running job.
    job_status = delegate.job_status(job_id)
    return JobStatusResponse(
        job_id=job_id,
        status_id=job_status.status_id,
        status_msg=job_status.status_text,
        is_complete=job_status.is_done,
        has_failed=job_status.has_failed
    ).json(), 200

@results_endpoint.route("/average_ensemble", methods=["POST"])
def make_average_ensemble():
    # Attempt to parse incoming request data.
    try:
        ensemble_request = AverageEnsembleRequest.parse_raw(request.json)

    except ValidationError as e:
        return ErrorResponse(msg=f"Invalid request data: '{e}'").json(), 400

    # Make sure we can grab a result with this ID.
    result_id = ensemble_request.result_id
    job_id = f"{result_id}-average_ensemble"

    # Ensure that the parent job (result_id) exists or has saved results.
    if not delegate.job_exists(result_id) and not delegate.job_complete(result_id):
        return ErrorResponse(msg=f"No parent job with ID '{result_id}' exists.").json(), 400

    # If a job or results object does not exist for this ID, start it.
    if not delegate.job_exists(job_id) and not delegate.job_complete(job_id):
        def results_to_json(results_json):
            results = Results.from_json(results_json)
            return results.average_ensemble().to_json()

        delegate.start_job(job_id, results_to_json, f"result://{result_id}")

    # Return the status of the currently running job.
    job_status = delegate.job_status(job_id)
    return JobStatusResponse(
        job_id=job_id,
        status_id=job_status.status_id,
        status_msg=job_status.status_text,
        is_complete=job_status.is_done,
        has_failed=job_status.has_failed
    ).json(), 200
